from ibm_watson_machine_learning.helpers.connections.connections import (
    DataConnection, S3Connection, S3Location, FSLocation, DSLocation, CP4DAssetLocation,
    WMLSAssetLocation, CloudAssetLocation, WSDAssetLocation, DeploymentOutputAssetLocation)
